/*
	solitairecrypt.c
	Copyright (C) 2001 David Fifield

	This file is part of solitairecrypt.

	solitairecrypt is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2, or (at
	your option) any later version.

	solitairecrypt is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <config.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <getopt.h>
#include <errno.h>
#include <sys/stat.h>

#include "solitairecrypt.h"
#include "solitaire.h"
#include "deck.h"

#if ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define _(s) gettext(s)
#else
# define _(s) s
# define textdomain(d)
# define bindtextdomain(p, d)
# define setlocale(a, b)
#endif

static char *prog_name;

int main(int argc, char *argv[])
{
	struct prog_option_t prog_option = {
		NULL,
		5,
		80,
		SOLITAIRE_ENCRYPT,
		"",
		NULL,
	};
	char (*solitaire_crypt[])(deck *d, char c) = {
		solitaire_encrypt,
		solitaire_decrypt
	};
	FILE *in_file, *out_file;
	struct stat in_stat, out_stat;
	deck d;
	int b, l;
	int c;

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	set_prog_option(argc, argv, &prog_option);

	deck_init(&d);
	solitaire_key(&d, prog_option.keyphrase);
	memset(prog_option.keyphrase, 0, strlen(prog_option.keyphrase));

	if ((prog_option.in_filename != NULL)
	    && (prog_option.out_filename != NULL)) {
		if (stat(prog_option.in_filename, &in_stat) == -1)
			error("%s: %s", prog_option.in_filename, strerror(errno));
		if (stat(prog_option.out_filename, &out_stat) != -1) {
			if (in_stat.st_ino == out_stat.st_ino)
				error(_("`%s' and `%s' are the same file"),
				      prog_option.in_filename, prog_option.out_filename);
		}
	}

	if (prog_option.in_filename == NULL)
		in_file = stdin;
	else
		in_file = fopen(prog_option.in_filename, "rb");
	if (in_file == NULL)
		error("%s: %s", prog_option.in_filename, strerror(errno));

	if (prog_option.out_filename == NULL)
		out_file = stdout;
	else
		out_file = fopen(prog_option.out_filename, "w");
	if (out_file == NULL)
		error("%s: %s", prog_option.out_filename, strerror(errno));

	b = 0;
	l = 0;
	while ((c = getc(in_file)) != EOF) {
		if (isalpha(c) == 0)
			continue;
		if (b >= prog_option.block_size) {
			b = 0;
			if (prog_option.line_len > 0
			    && prog_option.line_len <= l + prog_option.block_size) {
				putc('\n', out_file);
				l = 0;
			} else if (prog_option.block_size > 0) {
				putc(' ', out_file);
				l++;
			}
		}
		putc(solitaire_crypt[prog_option.crypt_dir](&d, c), out_file);
		b++;
		l++;
	}
	if (b > 0 && prog_option.crypt_dir == 0) {
		for ( ; b < prog_option.block_size; b++)
			putc(solitaire_crypt[prog_option.crypt_dir](&d, 'X'), out_file);
		l += prog_option.block_size + 1;
	}
	if (prog_option.line_len > 0)
		putc('\n', out_file);

	fflush(NULL);
	fclose(in_file);
	fclose(out_file);

	return 0;
}

void set_prog_option(int argc, char *argv[],
                     struct prog_option_t *prog_option)
{
	static struct option long_options[] = {
		{ "blocks", required_argument, NULL, 'b' },
		{ "columns", required_argument, NULL, 'c' },
		{ "decrypt", no_argument, NULL, 'd' },
		{ "encrypt", no_argument, NULL, 'e' },
		{ "key", required_argument, NULL, 'k' },
		{ "output", required_argument, NULL, 'o' },
		{ "help", no_argument, NULL, 0 },
		{ "version", no_argument, NULL, 0 }
	};
	int option_index;
	char c;

	prog_name = argv[0];

	option_index = 0;
	while ((c = getopt_long(argc, argv, "b:c:dek:o:",
	                        long_options, &option_index)) != -1) {
		switch (c) {
			case 'b':	/* blocks */
				if (sscanf(optarg, "%u", &prog_option->block_size) == 0)
					error(_("bad block size: %s"), optarg);
				break;
			case 'c':	/* columns */
				if (sscanf(optarg, "%u", &prog_option->line_len) == 0)
					error(_("bad number of columns: %s"), optarg);
				break;
			case 'd':	/* decrypt */
				prog_option->crypt_dir = SOLITAIRE_DECRYPT;
				break;
			case 'e':	/* encrypt */
				prog_option->crypt_dir = SOLITAIRE_ENCRYPT;
				break;
			case 'k':	/* key */
				prog_option->keyphrase = optarg;
				break;
			case 'o':
				prog_option->out_filename = optarg;
				break;
			case 0:
				switch (option_index) {
					case 6:	/* help */
						print_help(stdout);
						exit(0);
						break;
					case 7:	/* version */
						print_version(stdout);
						exit(0);
						break;
					default:	/* ??? */
						error(_("\
Unrecognized long option: %s\n\
This can't happen."), long_options[option_index].name);
						break;
				}
				break;
			case '?':
				fprintf(stderr, _("\
Try `%s --help' for more information.\n"), PACKAGE);
				exit(1);
			default:
				error(_("this can't happen."));
				break;
		}
	}
	if (optind < argc) {
		if (strcmp(argv[optind], "-") != 0)
			prog_option->in_filename = argv[optind];
		optind++;
		if (optind < argc)
			error(_("you may specify only one input file"));
	}

	return;
}

void print_help(FILE *s)
{
	fprintf(s, _("\
Usage: %s [OPTION]... [FILE]\n\
solitairecrypt encrypts or decrypts FILE and prints the resulting\n\
ciphertext or plaintext to a file or to standard output. Encryption is\n\
the default.\n\
\n\
  -b, --blocks=NUMBER    arrange output in blocks of NUMBER characters\n\
                           (default 5)\n\
  -c, --columns=NUMBER   arrange output in lines of NUMBER columns\n\
                           (default 80)\n\
  -d, --decrypt          specify decryption\n\
  -e, --encrypt          specify encryption (default)\n\
  -k, --key=STRING       key the deck using STRING as a keyphrase\n\
  -o, --output=FILE      write output to FILE. Without this option,\n\
                           write to standard output.\n\
  --help                 display a usage message and exit\n\
  --version              display version information and exit\n\
\n\
If no FILE is given or if FILE is `-', read standard input.\n\
\n\
Report bugs to <david@bamsoftware.com>.\n\
"), PACKAGE);

	return;
}

void print_version(FILE *s)
{
	fprintf(s, "%s %s\n", PACKAGE, VERSION);

	return;
}

void error(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", prog_name);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	va_end(ap);
	exit(1);
}

void warning(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, _("%s: warning: "), prog_name);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	va_end(ap);

	return;
}

